from inspect import Parameter, signature
from argparse import ArgumentParser, BooleanOptionalAction, Action, RawDescriptionHelpFormatter
from pathlib import Path
from typing import Any, Callable, NamedTuple, Sequence, Self
from textwrap import wrap, dedent

from erdb.table import Table
from erdb.loaders import GAME_VERSIONS
from erdb.utils.changelog import FormatterBase
from erdb.typing.game_version import GameVersion, GameVersionRange


def _parse_all_tables(tables: list[Table], all_effective_tables) -> list[Table]:
    table_set = set(tables)

    if Table.ALL in table_set:
        table_set.update(all_effective_tables)
        table_set.remove(Table.ALL)

    return list(table_set)

def _infer_annotation(type: Any, **kwargs) -> Any:
    if kwargs.get("action") in ["store_true", BooleanOptionalAction]:
        return bool
    
    assert type is not None, "Annotation cannot be deduced if no type is provided."

    if "default" in kwargs:
        if kwargs["default"] == None:
            return type | None

        if isinstance(kwargs["default"], list):
            return list[type]

    return type

class _TablesAction(Action):
    def __call__(self, parser, namespace, values: list[Table], option_string=None):
        setattr(namespace, self.dest, _parse_all_tables(values, Table))

class _GamedataAction(Action):
    def __call__(self, parser, namespace, values: list[str], option_string=None):
        setattr(namespace, self.dest, GameVersionRange.from_string(" ".join(values)))

class _ItemTypesAction(Action):
    def __call__(self, parser, namespace, values: list[Table], option_string=None):
        setattr(namespace, self.dest, _parse_all_tables(values, _ItemTypesAction.choices()))

    @staticmethod
    def choices() -> list[Table]:
        return [Table.ALL] + [tb for tb in Table.effective() if tb.spec.has_icons()]

class _Argument(NamedTuple):
    names: tuple
    param: Parameter
    kwargs: dict

    @classmethod
    def make(cls, *names: str, type: Any | None = None, annotation: Any | None = None, **kwargs) -> Self:
        assert len(names) > 0, "No names given to _Argument."
        param_name = names[0].removeprefix("--").replace("-", "_")

        if annotation is None:
            annotation = _infer_annotation(type, **kwargs)

        return cls(
            names,
            Parameter(param_name, Parameter.POSITIONAL_OR_KEYWORD, annotation=annotation),
            kwargs | ({} if type is None else {"type": type}) 
        )

    @classmethod
    def outputs_file(cls) -> list[Self]:
        return [
            cls.make("--out", "-o", type=Path, default=None, help="Optional output path.")
        ]

    @classmethod
    def outputs_json(cls) -> list[Self]:
        return [
            cls.make("--minimize", action="store_true", help="Output minimized JSON when generating data.")
        ] + cls.outputs_file()

    @classmethod
    def parses_gamedata(cls) -> list[Self]:
        default = GameVersionRange.from_version(GAME_VERSIONS[0]) if len(GAME_VERSIONS) > 0 else None
        return [
            cls.make(
                "--gamedata", "-g", annotation=GameVersionRange, default=default, nargs="+",
                action=_GamedataAction, help="Game version range to source the data from."
            )
        ]

    @classmethod
    def sources_gamedata(cls) -> list[Self]:
        return [
            cls.make("--game-dir", type=Path, required=True, help="Path to ELDEN RING's \"Game\" directory, where the binary is located."),
            cls.make("--ignore-checksum", action=BooleanOptionalAction, help="Ignore MD5 verification of thirdparty tools."),
            cls.make("--keep-cache", action=BooleanOptionalAction, help="Keep the unpacked files, if any."),
        ]

class _Subcommand(NamedTuple):
    command: str
    summary: str
    details: str
    aliases: list[str] = []
    examples: list[tuple[str, str]] = []
    arguments: list[_Argument] = []

    @classmethod
    def iterate(cls):
        yield from iter(cls.__subclasses__())

    @classmethod
    def get_parameters(cls) -> list[Parameter]:
        return [arg.param for arg in cls.arguments]

    @classmethod
    def get_description(cls) -> str:
        return "\n".join(wrap(dedent(cls.details), break_on_hyphens=False))

    @classmethod
    def get_epilog(cls) -> str:
        to_str = lambda x: f"[EXAMPLE] {x[0]}\n$ {x[1]}"
        return "\n\n".join(map(to_str, cls.examples))

class Generate(_Subcommand):
    command = "generate"
    summary = "Generate JSON data for specified tables."
    details = """\
    Parse extracted gamedata into a well-structured JSON output.
    The resulting data will be written to `{table}.json` files in a folder named after the `--out` argument, or cwd if not provided.
    This is a manual way of generating data, many other subcommands do this automatically.
    """

    aliases = ["gen"]

    examples = [
        (
            "Generate Armor data only for version 1.07.0",
            "erdb gen armor --gamedata only 1.07.0",
        ),
        (
            "Generate Talisman and Tool data for versions between 1.04.1 and 1.07.1 (including)",
            "erdb gen talismans tools --gamedata from 1.04.1 until 1.07.1",
        ),
        (
            "Generate all data for all versions from 1.06.0 (including), minimize the JSON output",
            "erdb gen all --gamedata until 1.06.0 --minimize",
        ),
    ]

    arguments = [
        _Argument.make("tables", type=Table, default=[], choices=list(Table), nargs="+", action=_TablesAction, help="Specify any or all tables.")
    ] + _Argument.parses_gamedata() + _Argument.outputs_json()

class FindValues(_Subcommand):
    command = "find-values"
    summary = "Find all possible values of a field per param name."
    details = """\
    Search for all possible values of a particular property in the specified game param.
    Unique values will be listed on each line, with example item/row names, or only their IDs if the name is blank.
    Useful for figuring out which properties are important and how the rows are grouped in relation to them.
    """

    aliases = ["vals"]

    examples = [
        (
            "Show unique values for \"sortGroupId\" property of \"EquipParamGoods\"",
            "erdb vals EquipParamGoods sortGroupId",
        ),
        (
            "Show weapon categories in \"EquipParamWeapon\", with up to 20 examples per category",
            "erdb vals EquipParamWeapon weaponCategory --limit 20",
        ),
    ]

    arguments = [
        _Argument.make("param", type=str, help="Name of the parameter table, ex. SpEffectParam"),
        _Argument.make("field", type=str, help="Name of the field of the table, ex. slashDamageCutRate"),
        _Argument.make("--limit", "-l", type=int, default=8, metavar="NUM", help="Limit of examples shown for each value (default 8)."),
    ] + _Argument.parses_gamedata()

class CalculateAR(_Subcommand):
    command = "calculate-ar"
    summary = "Calculate attack power of an armament"
    details = """\
    Calculate attack power and status effects of an armament of any affinity or level combination using the generated data.
    Provide the path to the data output directory from `erdb generate` via `--data-path`.
    Somber weapon levels are as-is, you must know the maximum level of the given weapon or wether the affinity is interchangable.
    Providing a level higher than maxiumum or an invalid affinity will result in an error.
    """

    aliases = ["ar"]

    examples = [
        (
            "Get AR and status effects for Poison Claymore +25 at 25 str, 30 dex, 10 int, 10 fth, 20 arc",
            "erdb ar 25,30,10,10,20 Claymore Poison 25 --data-path /path/to/generated/data/1.07.1/",
        ),
        (
            "Get AR for Standard Axe of Godfrey +8 at 50 str, 20 dex, 10 int, 10 fth, 10 arc",
            "erdb ar 50,20,10,10,10 \"Axe of Godfrey\" Standard 8 --data-path ./1.06.0",
        ),
    ]

    arguments = [
        _Argument.make("attribs", type=str, help="Player attributes in format \"str,dex,int,fth,arc\"."),
        _Argument.make("armament", type=str, help="Name of the armament."),
        _Argument.make("affinity", type=str, help="Affinity of the armament."),
        _Argument.make("level", type=int, help="Upgrade level of the armament."),
        _Argument.make("--data-path", type=Path, required=True, help="Location of the generated data."),
    ]

class Changelog(_Subcommand):
    command = "changelog"
    summary = "Create a changelog of ERDB-detectable differences between specified versions."
    details = """\
    Compare two generated data and create a human-readable list of updates.
    ERDB automatically finds the previous version from the one requested, but that can be overridden with an optional `--from-version` argument.
    Different formatters are available, using which the output will be written to stdout, or a specific file if `--out` is provided.
    """

    aliases = []

    examples = [
        (
            "Generate changelog from version 1.06.0 to 1.07.0 and write it to CHANGELOG.md using markdown format",
            "erdb changelog 1.07.0 --formatter markdown --out CHANGELOG.md",
        ),
        (
            "Print changelog from version 1.05.0 to 1.07.1 in a text format to stdout",
            "erdb changelog 1.07.1 --from-version 1.05.0 --formatter text",
        ),
    ]

    arguments = [
        _Argument.make("version", type=GameVersion.from_string, annotation=GameVersion, help="Version to generate the changelog of."),
        _Argument.make("--from-version", type=GameVersion.from_string, annotation=GameVersion | None, default=None, help="Optional starting version of the changelog, previous if not specified."),
        _Argument.make("--formatter", "-f", type=str, default=FormatterBase.identifiers()[0], choices=FormatterBase.identifiers(), help="Format to output the changelog in."),
    ] + _Argument.outputs_file()

class Source(_Subcommand):
    command = "source"
    summary = "Extract gamedata from an UXM-unpacked ELDEN RING installation (Windows only)."
    details = """\
    Parse an UXM-unpacked installation of ELDEN RING and collect data to ERDB package resources.
    In order to extract the files beyond UXM, third-party tools will be downloaded, verified and executed.
    The version of the game is automatically detected but can be overridden for any purpose, ex. modded version.
    This version can then be used in other ERDB subcommands, like generating tables or calculating AR.
    """

    aliases = []

    examples = [
        (
            "Source a game folder and autodetect its version",
            "erdb source --game-dir \"/path/to/ELDEN RING/Game\"",
        ),
        (
            "Source a game folder, specifying custom version and keep whatever was additionally extracted",
            "erdb source --game-dir \"/path/to/ELDEN RING/Game\" --keep-cache --version 1.99.0",
        ),
    ]

    arguments = [
        _Argument.make("--version", "-v", type=GameVersion.from_string, annotation=GameVersion | None, default=None, help="Version directory storing the extracted files, overrides autodetection."),
    ] + _Argument.sources_gamedata()

class Map(_Subcommand):
    command = "map"
    summary = "Extract world map image from an UXM-unpacked ELDEN RING installation (Windows only)."
    details = """\
    Parse an UXM-unpacked installation of ELDEN RING and extract the world map tiles.
    Tiles are combined into an image and written to the specified file or directory using the `--out` argument.
    Provide a file extension to automatically convert it to that format.
    """

    aliases = []

    examples = [
        (
            "Write the full quality map to er_map.jpeg",
            "erdb map --game-dir \"/path/to/ELDEN RING/Game\" -o er_map.jpeg",
        ),
        (
            "Write the LOD 2 quality map of underground to a default filename and extension",
            "erdb map --game-dir \"/path/to/ELDEN RING/Game\" --lod 2 --underground",
        ),
    ]

    arguments = [
        _Argument.make("--lod", "-l", type=int, default=0, help="Level of detail of the map, 0 is highest."),
        _Argument.make("--underground", action=BooleanOptionalAction, help="Specifies whether to extract the underground map instead of the overworld."),
    ] + _Argument.sources_gamedata() + _Argument.outputs_file()

class Icons(_Subcommand):
    command = "icons"
    summary = "Extract item images from an UXM-unpacked ELDEN RING installation (Windows only)."
    details = """\
    Parse an UXM-unpacked installation of ELDEN RING and extract the item icons.
    Accepts list of tables similar to `erdb generate` but only ones that actually are in-game items.
    Item icons are output in directories with the same name as their respective table, parented by the `--out` argument, or cwd if not provided.
    `--file-format` can be used to name them in a custom way and provide a specific extension to automatically format images to.
    """

    aliases = []

    examples = [
        (
            "Extract icons for tools and keys in specific file names and as ICO format",
            "erdb icons tools keys --game-dir \"/path/to/ELDEN RING/Game\" --file-format \"{icon_id} - {name}.ico\"",
        ),
        (
            "Extract all icons in a small file size",
            "erdb icons all --game-dir \"/path/to/ELDEN RING/Game\" --size 128",
        ),
    ]

    arguments = [
        _Argument.make("types", type=Table, default=[], nargs="+", choices=_ItemTypesAction.choices(), action=_ItemTypesAction, help="Specify item types to export images for."),
        _Argument.make("--size", "-s", type=int, default=1024, choices=range(1, 1025), metavar="[1-1024]", help="Size in pixels of images to be exported, resized from maximum quality in game files (1024x1024)."),
        _Argument.make("--file-format", "-f", type=str, default="{icon_id}.png", help="Specify the formatting for file names, including extension. Available fields: {icon_id}, {name}, {table}. NOTE: multiple items can share {icon_id}, therefore {name} alone is not exhaustive and only one will be used."),
    ] + _Argument.sources_gamedata() + _Argument.outputs_file()

class ServeAPI(_Subcommand):
    command = "serve-api"
    summary = "Begin serving the API web server."
    details = """\
    Start a web server providing a REST API for every table and for all available game versions.
    Data is served from memory, and is generated lazily unless `--precache` is provided.
    The full documentation is available under `/v{api_version}/docs` or `/v{api_version}/redoc` endpoints.
    Higher level endpoints, `/docs` and `/redoc` outline all API versions there are thus far.
    """

    aliases = ["api"]

    examples = [
        (
            "Serve the REST API on port 8080 and pracache all the tables",
            "erdb serve-api --port 8080 --precache",
        ),
    ]

    arguments = [
        _Argument.make("--port", "-p", type=int, required=True, help="Port number to listen on."),
        _Argument.make("--bind", "-b", type=str, default="0.0.0.0", help="Address to bind the server on."),
        _Argument.make("--precache", action=BooleanOptionalAction, help="Pregenerate all data instead of lazy loading."),
    ]

def parse_args(argv: Sequence[str], handlers: dict[str, Callable]) -> dict[str, Any]:
    parser = ArgumentParser(description="Interface for ERDB operations.")
    subs = parser.add_subparsers(title="subcommands", required=True)

    for cmd in _Subcommand.iterate():
        assert cmd.command in handlers, f"No handler found for \"{cmd.command}\" subcommand"
        handler = handlers[cmd.command]

        sig = signature(handler)
        assert sig.return_annotation == int, f"Handler for \"{cmd.command}\" must return an int"
        assert cmd.get_parameters() == list(sig.parameters.values()), f"Parameter mismatch for \"{cmd.command}\" handler"

        p = subs.add_parser(cmd.command, help=cmd.summary, description=cmd.get_description(), aliases=cmd.aliases, epilog=cmd.get_epilog(), formatter_class=RawDescriptionHelpFormatter)

        for arg in cmd.arguments:
            p.add_argument(*arg.names, **arg.kwargs)

        p.set_defaults(handler=handler)

    return vars(parser.parse_args(argv))


from pydantic import Field
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config
from erdb.typing.models.armament import Affinity
from erdb.typing.models.item import Item
from erdb.typing.categories import ArmamentCategory


@dataclass(config=dt_config())
class AshOfWar(Item):
    armament_categories: list[ArmamentCategory] = Field(...,
        description="Applicable weapons classes the Ash of War can be applied to.",
        min_items=1,
        example=[ArmamentCategory.GREATAXE, ArmamentCategory.GREAT_HAMMER, ArmamentCategory.COLOSSAL_WEAPON],
    )
    default_affinity: Affinity = Field(...,
        description="Default Affinity the Ash of War comes with.",
        example=Affinity.FLAME_ART,
    )
    possible_affinities: list[Affinity] = Field(...,
        description="List of Affinities the Ash of War can provide, assuming Whetblades are available.",
        min_items=1,
        example=[Affinity.STANDARD, Affinity.HEAVY, Affinity.KEEN, Affinity.QUALITY, Affinity.FIRE, Affinity.FLAME_ART]
    )
    skill_id: int = Field(...,
        description="Index of the Skill the Ash of War comes with.",
        ge=10,
        example=113,
    )



from erdb.typing.models.reinforcement import Reinforcement, ReinforcementLevel, DamageMultiplier, ScalingMultiplier, GuardMultiplier, ResistanceMultiplier
from erdb.typing.params import ParamRow
from erdb.typing.enums import ItemIDFlag
from erdb.typing.api_version import ApiVersion
from erdb.utils.common import find_offset_indices
from erdb.table._retrievers import ParamDictRetriever, RetrieverData
from erdb.table._common import RowPredicate, TableSpecContext


def _get_damages(row: ParamRow) -> DamageMultiplier:
    return DamageMultiplier(
        physical=row["physicsAtkRate"].as_float,
        magic=row["magicAtkRate"].as_float,
        fire=row["fireAtkRate"].as_float,
        lightning=row["thunderAtkRate"].as_float,
        holy=row["darkAtkRate"].as_float,
        stamina=row["staminaAtkRate"].as_float,
    )

def _get_scalings(row: ParamRow) -> ScalingMultiplier:
    return ScalingMultiplier(
        strength=row["correctStrengthRate"].as_float,
        dexterity=row["correctAgilityRate"].as_float,
        intelligence=row["correctMagicRate"].as_float,
        faith=row["correctFaithRate"].as_float,
        arcane=row["correctLuckRate"].as_float,
    )

def _get_guards(row: ParamRow) -> GuardMultiplier:
    return GuardMultiplier(
        physical=row["physicsGuardCutRate"].as_float,
        magic=row["magicGuardCutRate"].as_float,
        fire=row["fireGuardCutRate"].as_float,
        lightning=row["thunderGuardCutRate"].as_float,
        holy=row["darkGuardCutRate"].as_float,
        guard_boost=row["staminaGuardDefRate"].as_float,
    )

def _get_resistances(row: ParamRow) -> ResistanceMultiplier:
    return ResistanceMultiplier(
        poison=row["poisonGuardResistRate"].as_float,
        scarlet_rot=row["diseaseGuardResistRate"].as_float,
        frostbite=row["freezeGuardDefRate"].as_float,
        bleed=row["bloodGuardResistRate"].as_float,
        sleep=row["sleepGuardDefRate"].as_float,
        madness=row["madnessGuardDefRate"].as_float,
        death_blight=row["curseGuardResistRate"].as_float,
    )

def _get_reinforcement_level(row: ParamRow, level: int) -> ReinforcementLevel:
    return ReinforcementLevel(
        level=level,
        damage=_get_damages(row),
        scaling=_get_scalings(row),
        guard=_get_guards(row),
        resistance=_get_resistances(row)
    )

class ReinforcementTableSpec(TableSpecContext):
    model = {
        ApiVersion.VER_1: Reinforcement,
    }

    main_param_retriever = ParamDictRetriever("ReinforceParamWeapon", ItemIDFlag.NON_EQUIPABBLE)

    predicates: list[RowPredicate] = [
        lambda row: row.is_base_item,
        lambda row: len(row.name) > 0,
    ]

    @classmethod # override
    def get_pk(cls, data: RetrieverData, row: ParamRow) -> str:
        return str(row.index)

    @classmethod
    def make_object(cls, api: ApiVersion, data: RetrieverData, row: ParamRow):
        indices, offset = find_offset_indices(row.index, data.main_param, possible_maxima=[0, 10, 25])
        return Reinforcement([_get_reinforcement_level(data.main_param[i], lvl) for i, lvl in zip(indices, offset)])


from typing import Literal
from pydantic import Field, PositiveInt
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config, NonEmptyStr
from erdb.typing.enums import Region, Location, Currency


@dataclass(config=dt_config())
class LocationDetail:
    summary: str = Field("no summary",
        description="Short, consice summary of the location. To help concatenating with other data, there are no capital letters or periods at the end.",
        min_length=1, regex=r"^.+(?<!\\.)$",
        example="found on top a giant spear in Leyndell, Royal Capital",
    )
    quantity: PositiveInt | Literal["infinite"] | None = Field(None,
        description="Specifies the amount if an integer, otherwise `infinite` if the Item respawns or can be purchased infinitely.",
        example=1,
    )
    location: Location | None = Field(None,
        description="The specific location in which the Item is found.",
        example=Location.LEYNDELL_ROYAL_CAPITAL,
    )
    region: Region | None = Field(None,
        description="The generic region in which the Item is found.",
        example=Region.ALTUS_PLATEAU,
    )
    directions: NonEmptyStr | None =  Field(None,
        description="Exact description on where to find the Item if summary cannot be straightfoward enough.",
        example="From the East Capital Rampart Gate grace, take an elevator down towards the capital. Head alongside the railing to the left and jump onto the giant spear.",
    )
    price_bought: PositiveInt | None = Field(None,
        description="The amount of Currency the Item is bought for at this location, if applicable.",
        example=500,
    )
    currency: Currency | None =  Field(None,
        description="The type of currency this item is bought for, if applicable.",
        example=Currency.RUNES,
    )
    requirements: list[NonEmptyStr] | None = Field(None,
        description="List of requirements which make the item available in full sentences.",
        min_items=1,
        example=[
            "Two Great Runes must be acquired to access Leyndell, Royal Capital",
            "Godfrey, First ELden Lord (Golden Shade) must be defeated to gain access to the area with the Armament.",
        ],
    )
    blockers: list[NonEmptyStr] | None = Field(None,
        description="List of situations which cause the item to become unavailable in full sentences.",
        min_items=1,
        example=["Maliketh, the Black Blade is deafeated turning Leyndell, Royal Capital into Leyndell, Ashen Capital."]
    )


from erdb.typing.models.ash_of_war import AshOfWar
from erdb.typing.params import ParamRow
from erdb.typing.enums import ItemIDFlag, Affinity
from erdb.typing.categories import ArmamentCategory
from erdb.typing.api_version import ApiVersion
from erdb.table._retrievers import ParamDictRetriever, MsgsRetriever, RetrieverData
from erdb.table._common import TableSpecContext


def _get_categories(row: ParamRow) -> list[ArmamentCategory]:
    return [a for a in list(ArmamentCategory) if row[f"canMountWep_{a.ingame}"].as_bool]

def _get_affinities(row: ParamRow) -> list[Affinity]:
    return [a for a in Affinity if row[f"configurableWepAttr{str(a.id).zfill(2)}"].as_bool]

class AshOfWarTableSpec(TableSpecContext):
    model = {
        ApiVersion.VER_1: AshOfWar,
    }

    main_param_retriever = ParamDictRetriever("EquipParamGem", ItemIDFlag.ACCESSORIES, id_min=10000)

    msg_retrievers = {
        "names": MsgsRetriever("GemName"),
        "summaries": MsgsRetriever("GemInfo"),
        "descriptions": MsgsRetriever("GemCaption")
    }

    @classmethod
    def make_object(cls, api: ApiVersion, data: RetrieverData, row: ParamRow):
        return AshOfWar(
            **cls.make_item(data, row),
            **cls.make_contrib(data, row, "locations", "remarks"),
            armament_categories=_get_categories(row),
            default_affinity=Affinity.from_id(row["defaultWepAttr"].as_int),
            possible_affinities=_get_affinities(row),
            skill_id=row["swordArtsParamId"].as_int,
        )


from pydantic import Field
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config
from erdb.typing.models.item import Item
from erdb.typing.categories import BolsteringMaterialCategory


@dataclass(config=dt_config())
class BolsteringMaterial(Item):
    category: BolsteringMaterialCategory = Field(...,
        description="Bolstering Material category to discern its use.",
        example=BolsteringMaterialCategory.SMITHING_STONE,
    )


from copy import deepcopy
from enum import Enum
from types import SimpleNamespace
from typing import Any, Callable, NamedTuple, Self


class EffectType(str, Enum):
    POSITIVE = "positive"
    NEUTRAL = "neutral"
    NEGATIVE = "negative"

class EffectModel(str, Enum):
    MULTIPLICATIVE = "multiplicative"
    ADDITIVE = "additive"

"""
Helper class of attributes used for effects, these are self-made
and don't correspond to anything in the game.
"""
class AttributeName(str, Enum):
    MAXIMUM_HEALTH = "Maximum Health",
    HEALTH_POINTS = "Health Points",
    FLASK_HEALTH_RESTORATION = "Flask Health Restoration",
    MAXIMUM_FOCUS = "Maximum Focus",
    FOCUS_POINTS = "Focus Points",
    FLASK_FOCUS_RESTORATION = "Flask Focus Restoration",
    MAXIMUM_STAMINA = "Maximum Stamina",
    STAMINA_RECOVERY_SPEED = "Stamina Recovery Speed",
    MAXIMUM_EQUIP_LOAD = "Maximum Equip Load",
    POISE = "Poise",
    VIGOR = "Vigor",
    MIND = "Mind",
    ENDURANCE = "Endurance",
    STRENGHT = "Strength",
    DEXTERITY = "Dexterity",
    INTELLIGENCE = "Intelligence",
    FAITH = "Faith",
    ARCANE = "Arcane",
    STANDARD_ABSORPTION = "Standard Absorption",
    STRIKE_ABSORPTION = "Strike Absorption",
    SLASH_ABSORPTION = "Slash Absorption",
    PIERCE_ABSORPTION = "Pierce Absorption",
    PHYSICAL_ABSORPTION = "Physical Absorption",
    MAGIC_ABSORPTION = "Magic Absorption",
    FIRE_ABSORPTION = "Fire Absorption",
    LIGHTNING_ABSORPTION = "Lightning Absorption",
    HOLY_ABSORPTION = "Holy Absorption",
    ELEMENTAL_ABSORPTION = "Elemental Absorption",
    ABSORPTION = "Absorption",
    STANDARD_ATTACK_POWER = "Standard Attack Power",
    STRIKE_ATTACK_POWER = "Strike Attack Power",
    SLASH_ATTACK_POWER = "Slash Attack Power",
    PIERCE_ATTACK_POWER = "Pierce Attack Power",
    PHYSICAL_ATTACK_POWER = "Physical Attack Power",
    MAGIC_ATTACK_POWER = "Magic Attack Power",
    FIRE_ATTACK_POWER = "Fire Attack Power",
    LIGHTNING_ATTACK_POWER = "Lightning Attack Power",
    HOLY_ATTACK_POWER = "Holy Attack Power",
    ELEMENTAL_ATTACK_POWER = "Elemental Attack Power",
    ATTACK_POWER = "Attack Power",
    STAMINA_ATTACK_RATE = "Stamina Attack Rate"
    STABILITY = "Stability"
    IMMUNITY = "Immunity",
    ROBUSTNESS = "Robustness"
    VITALITY = "Vitality",
    FOCUS = "Focus",
    POISON_RESISTANCE = "Poison Resistance",
    SCARLET_ROT_RESISTANCE = "Scarlet Rot Resistance",
    BLEED_RESISTANCE = "Bleed Resistance",
    FROSTBITE_RESISTANCE = "Frostbite Resistance",
    SLEEP_RESISTANCE = "Sleep Resistance",
    MADNESS_RESISTANCE = "Madness Resistance",
    DEATH_BLIGHT_RESISTANCE = "Death Blight Resistance",
    MEMORY_SLOTS = "Memory Slots",
    CASTING_SPEED = "Casting Speed",
    SPELL_DURATION = "Spell Duration",
    SORCERY_FOCUS_CONSUMPTION = "Sorcery Focus Consumption",
    INCANTATION_FOCUS_CONSUMPTION = "Incantation Focus Consumption",
    PYROMANCY_FOCUS_CONSUMPTION = "Pyromancy Focus Consumption",
    SPELL_FOCUS_CONSUMPTION = "Spell Focus Consumption",
    SKILL_FOCUS_CONSUMPTION = "Skill Focus Consumption",
    BOW_DISTANCE = "Bow Distance",
    ENEMY_HEARING = "Enemy Hearing",
    FALL_DAMAGE = "Fall Damage",
    ITEM_DISCOVERY = "Item Discovery",
    RUNE_ACQUISITION = "Rune Acquisition",
    INVISIBLE_AT_DISTANCE = "Invisible at Distance"
    REDUCE_HEADSHOT_IMPACT = "Reduce Headshot Impact"
    SWITCH_ANIMATION_GENDER = "Switch Animation Gender"
    APPEAR_AS_COOPERATOR = "Appear as Cooperator"
    APPEAR_AS_HOST = "Appear as Host"
    PRESERVE_RUNES_ON_DEATH = "Preserve Runes on Death"
    DESTROY_ITEM_ON_DEATH = "Destroy Item on Death"
    ATTRACT_ENEMY_AGGRESSION = "Attract Enemy Aggression"

class AttributeField(NamedTuple):
    attribute: AttributeName
    effect_model: EffectModel
    effect_type: EffectType
    parser: Callable
    conditions: list[str] | None
    default_value: Any

    def get_effective_type(self, value: Any):
        T = EffectType
        value_increase = value >= self.default_value
        return {
            T.POSITIVE: { True: T.POSITIVE, False: T.NEGATIVE, },
            T.NEGATIVE: { True: T.NEGATIVE, False: T.POSITIVE, },
            T.NEUTRAL:  { True: T.NEUTRAL,  False: T.NEUTRAL,  }
        }[self.effect_type][value_increase]

    @classmethod
    def create(cls, attribute: AttributeName, effect_model: EffectModel, effect_type: EffectType,
               parser: Callable, conditions: list[str] | None = None, default_value: Any | None = None) -> Self:

        def _default_value_from_model():
            return 1 if effect_model == EffectModel.MULTIPLICATIVE else 0

        default_value = _default_value_from_model() if default_value is None else default_value
        return cls(attribute, effect_model, effect_type, parser, conditions, default_value)

class SchemaEffect(SimpleNamespace):
    attribute: AttributeName
    conditions: list[str] | None = None
    tick_interval: float | None = None
    effect_model: EffectModel
    effect_type: EffectType
    value: float
    value_pvp: float | None = None

    def to_dict(self) -> dict:
        d = {
            "attribute": self.attribute,
            "model": self.effect_model,
            "type": self.effect_type,
            "value": self.value,
        }

        for prop in ["conditions", "tick_interval", "value_pvp"]:
            if getattr(self, prop) is not None:
                d[prop] = getattr(self, prop)

        return d

    def get_values_hash(self) -> int:
        conds = None if self.conditions is None else tuple(self.conditions)
        return hash((conds, self.tick_interval, self.effect_model, self.effect_type, self.value, self.value_pvp))

    def clone(self, new_attribute: AttributeName) -> Self:
        new_effect = deepcopy(self)
        new_effect.attribute = new_attribute
        return new_effect

    def __str__(self) -> str:
        conds      = "" if self.conditions is None else f" (under conditions {self.conditions})"
        sign_val   = "+" if self.value > 0 else "-"
        sign_model = "%" if self.effect_model == EffectModel.MULTIPLICATIVE else "+"
        val_pvp    = "" if self.value_pvp is None else f" ({self.value_pvp} PVP)"
        tick       = "" if self.tick_interval is None else f" on tick {self.tick_interval} s"
        return f"{sign_val}{self.value}{sign_model}{val_pvp} {self.attribute.value}{conds}{tick}"

    @classmethod
    def from_attribute_field(cls, value: float, attrib_field: AttributeField) -> Self:
        value = attrib_field.parser(value, attrib_field.effect_model)

        return cls(
            attribute=attrib_field.attribute,
            conditions=attrib_field.conditions,
            effect_model=attrib_field.effect_model,
            effect_type=attrib_field.get_effective_type(value),
            value=value)

    @classmethod
    def from_obj(cls, obj: Any) -> Self:
        return cls(
            attribute=AttributeName(getattr(obj, "attribute")),
            conditions=getattr(obj, "conditions", None),
            tick_interval=getattr(obj, "tick_interval"),
            effect_model=EffectModel(getattr(obj, "model")),
            effect_type=EffectType(getattr(obj, "type")),
            value=getattr(obj, "value", None),
            value_pvp=getattr(obj, "value_pvp", None))


from pydantic import Field
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config
from erdb.typing.models.item import Item
from erdb.typing.categories import InfoCategory


@dataclass(config=dt_config())
class Info(Item):
    category: InfoCategory = Field(...,
        description="Info Item category to discern its use.",
        example=InfoCategory.NOTE,
    )


from pydantic import Field, NonNegativeInt
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config
from erdb.typing.models.item import Item
from erdb.typing.models.effect import Effect
from erdb.typing.categories import ToolCategory
from erdb.typing.enums import ToolAvailability


@dataclass(config=dt_config())
class Tool(Item):
    category: ToolCategory = Field(...,
        description="Tool category to discern its use.",
        example=ToolCategory.UTILITY,
    )
    availability: ToolAvailability = Field(...,
        description="Defines whether the Tool is available always, only during singleplayer or only during multiplayer.",
        example=ToolAvailability.ALWAYS,
    )
    fp_cost: NonNegativeInt = Field(...,
        description="Cost of FP to use.",
        example=0,
    )
    is_consumed: bool = Field(...,
        description="Specifies whether the Tool is consumed on use.",
        example=True,
    )
    is_ladder_usable: bool = Field(...,
        description="Specifies whether the Tool is available on ladders.",
        example=False,
    )
    is_horseback_usable: bool = Field(...,
        description="Specifies whether the Tool is available on horseback.",
        example=True,
    )
    effects: list[Effect] = Field(...,
        description="Effects of the Tool.",
        # example provided by Effects model
    )


from pydantic import Field, PositiveFloat, NonNegativeInt
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config
from erdb.typing.models.item import Item
from erdb.typing.models.effect import Effect
from erdb.typing.categories import ArmorCategory


@dataclass(config=dt_config())
class Absorptions:
    physical: float
    strike: float
    slash: float
    pierce: float
    magic: float
    fire: float
    lightning: float
    holy: float

@dataclass(config=dt_config())
class Resistances:
    immunity: int
    robustness: int
    focus: int
    vitality: int
    poise: int

@dataclass(config=dt_config())
class Armor(Item):
    category: ArmorCategory = Field(...,
        description="Category of the Armor.",
        example=ArmorCategory.BODY,
    )
    altered: str = Field(...,
        description="Name of the altered (or reversed) armor piece, empty if unalterable.",
        example="Elden Lord Armor (Altered)",
    )
    weight: PositiveFloat = Field(...,
        description="Weight of the Armor.",
        example=9.2,
    )
    icon_fem: NonNegativeInt = Field(...,
        description="Icon ID to the female version of the Armor, `icon` field specifies the male version which is usually the same.",
        example=584,
    )
    absorptions: Absorptions = Field(...,
        description="Absorption values for the Armor.",
        example=Absorptions(
            physical=11.9, strike=10.9, slash=11.4, pierce=12.4,
            magic=8.8, fire=11.4, lightning=7.1, holy=8.,
        ),
    )
    resistances: Resistances = Field(...,
        description="Resistance values for the Armor.",
        example=Resistances(immunity=32, robustness=55, focus=18, vitality=21, poise=19),
    )
    effects: list[Effect] = Field(...,
        description="Additional effects of the Armor."
        # example provided by Effect model
    )


from pydantic import Field, NonNegativeFloat, conlist
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config


@dataclass(config=dt_config())
class DamageMultiplier:
    physical: NonNegativeFloat
    magic: NonNegativeFloat
    fire: NonNegativeFloat
    lightning: NonNegativeFloat
    holy: NonNegativeFloat
    stamina: NonNegativeFloat

@dataclass(config=dt_config())
class ScalingMultiplier:
    strength: NonNegativeFloat
    dexterity: NonNegativeFloat
    intelligence: NonNegativeFloat
    faith: NonNegativeFloat
    arcane: NonNegativeFloat

@dataclass(config=dt_config())
class GuardMultiplier:
    physical: NonNegativeFloat
    magic: NonNegativeFloat
    fire: NonNegativeFloat
    lightning: NonNegativeFloat
    holy: NonNegativeFloat
    guard_boost: NonNegativeFloat

@dataclass(config=dt_config())
class ResistanceMultiplier:
    bleed: NonNegativeFloat
    frostbite: NonNegativeFloat
    poison: NonNegativeFloat
    scarlet_rot: NonNegativeFloat
    sleep: NonNegativeFloat
    madness: NonNegativeFloat
    death_blight: NonNegativeFloat

@dataclass(config=dt_config())
class ReinforcementLevel:
    level: int = Field(..., ge=0, le=25)
    damage: DamageMultiplier = Field(...)
    scaling: ScalingMultiplier = Field(...)
    guard: GuardMultiplier = Field(...)
    resistance: ResistanceMultiplier = Field(...)

Reinforcement = conlist(ReinforcementLevel, min_items=1, max_items=26)


from pydantic import Field
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config
from erdb.typing.models.item import Item
from erdb.typing.categories import ShopCategory


@dataclass(config=dt_config())
class Shop(Item):
    category: ShopCategory = Field(...,
        description="Shop Item category to discern its use.",
        example=ShopCategory.COOKBOOK,
    )


from erdb.typing.models.spells import Spell
from erdb.typing.models.common import StatRequirements
from erdb.typing.params import ParamRow
from erdb.typing.enums import GoodsType, ItemIDFlag, SpellHoldAction
from erdb.typing.categories import SpellCategory
from erdb.typing.api_version import ApiVersion
from erdb.utils.common import remove_nulls
from erdb.table._retrievers import ParamDictRetriever, MsgsRetriever, RetrieverData
from erdb.table._common import RowPredicate, TableSpecContext


def _get_spell_requirements(row: ParamRow) -> StatRequirements:
    data = {
        "intelligence": row["requirementIntellect"].get_int(null_value=0),
        "faith": row["requirementFaith"].get_int(null_value=0),
        "arcane": row["requirementLuck"].get_int(null_value=0),
    }

    return StatRequirements(**remove_nulls(data))

class SpellTableSpec(TableSpecContext):
    model = {
        ApiVersion.VER_1: Spell,
    }

    # Spells are defined in Goods and Magic tables, correct full hex IDs are calculated from Goods
    main_param_retriever = ParamDictRetriever("EquipParamGoods", ItemIDFlag.GOODS)

    predicates: list[RowPredicate] = [
        lambda row: 1 <= row["sortId"].as_int < 999999,
        lambda row: row["goodsType"] in [GoodsType.SORCERY_1, GoodsType.INCANTATION_1, GoodsType.SORCERY_2, GoodsType.INCANTATION_2],
    ]

    param_retrievers = {
        "magic": ParamDictRetriever("Magic", ItemIDFlag.NON_EQUIPABBLE)
    }

    msg_retrievers = {
        "names": MsgsRetriever("GoodsName"),
        "summaries": MsgsRetriever("GoodsInfo"),
        "descriptions": MsgsRetriever("GoodsCaption")
    }

    @classmethod
    def make_object(cls, api: ApiVersion, data: RetrieverData, row: ParamRow):
        row_magic = data.params["magic"][row.index]

        fp_cost = row_magic["mp"].as_int
        fp_extra = row_magic["mp_charge"].as_int

        sp_cost = row_magic["stamina"].as_int
        sp_extra = row_magic["stamina_charge"].as_int

        hold_action = SpellHoldAction.NONE if fp_extra == 0 else SpellHoldAction.CHARGE
        hold_action = SpellHoldAction.CONTINUOUS if row_magic["consumeLoopMP_forMenu"].get_int() else hold_action 

        return Spell(
            **cls.make_item(data, row),
            **cls.make_contrib(data, row, "locations", "remarks"),
            fp_cost=fp_cost,
            fp_cost_extra=fp_extra - fp_cost if hold_action == "Charge" else fp_extra,
            sp_cost=sp_cost,
            sp_cost_extra=sp_extra - sp_cost if hold_action == "Charge" else sp_extra,
            category=SpellCategory.from_row(row_magic),
            slots_used=row_magic["slotLength"].as_int,
            hold_action=hold_action,
            is_weapon_buff=row_magic["isEnchant"].as_bool,
            is_shield_buff=row_magic["isShieldEnchant"].as_bool,
            is_horseback_castable=row_magic["enableRiding"].as_bool,
            requirements=_get_spell_requirements(row_magic),
        )


from typing import Any
from fastapi import Query, status
from fastapi.responses import JSONResponse
from pydantic.dataclasses import dataclass

from erdb.table import Table
from erdb.app_api.common import GameVersionEnum, generate
from erdb.utils.common import getattrstr
from erdb.typing.api_version import ApiVersion


@dataclass
class _Detail:
    detail: str

class DataEndpoint:
    api: ApiVersion
    table: Table

    def __init__(self, api: ApiVersion, table: Table) -> None:
        self.api = api
        self.table = table

    @property
    def route(self) -> str:
        return "/"

    @property
    def model(self) -> Any:
        return dict[str, self.table.spec.model[self.api]] # type: ignore

    @property
    def summary(self) -> str:
        return "multiple items"

    @property
    def description(self) -> str:
        return "Retrieve a dictionary of item's ascii name -> item properties."

    @property
    def responses(self) -> dict[int, dict]:
        return {
            status.HTTP_400_BAD_REQUEST: {
                "model": _Detail,
                "description": "Bad Request: `query` format is valid, but the specified field does not exist for this model."
            }
        }

    def __call__(self,
        game_version: GameVersionEnum,
        keys: list[str] | None = Query(None, alias="k", description="Specify a list of keys (ascii names) to retrieve specific items."),
        query: str | None = Query(None, description="Filter elements by field in format \"{field}:{value}\".", regex=r"^\w+\:.+$"),
    ) -> Any:
        data = generate(self.api, game_version, self.table)

        if keys is not None:
            data = {k: v for k, v in data.items() if k in keys}

        if query is not None:
            field, value = query.split(":", maxsplit=1)

            try:
                data = {k: v for k, v in data.items() if getattrstr(v, field) == value}
            
            except AttributeError:
                return JSONResponse(status_code=status.HTTP_400_BAD_REQUEST, content={"detail": f"{self.table.title} has no field: \"{field}\"."})

        return data

class ItemEndpoint:
    api: ApiVersion
    table: Table

    def __init__(self, api: ApiVersion, table: Table) -> None:
        self.api = api
        self.table = table

    @property
    def route(self) -> str:
        return "/{key}"

    @property
    def model(self) -> Any:
        return self.table.spec.model[self.api]

    @property
    def summary(self) -> str:
        return "single item"

    @property
    def description(self) -> str:
        return "Retrieve properties of a single item."

    @property
    def responses(self) -> dict[int, dict]:
        return {
            status.HTTP_404_NOT_FOUND: {"model": _Detail}
        }

    def __call__(self, game_version: GameVersionEnum, key: str) -> Any:
        try:
            return generate(self.api, game_version, self.table)[key]

        except KeyError:
            return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content={"detail": f"{self.table.title} has no key: \"{key}\""})


from functools import cache
from enum import Enum

from erdb.table import Table
from erdb.loaders import GAME_VERSIONS
from erdb.typing.game_version import GameVersion
from erdb.typing.api_version import ApiVersion


GameVersionEnum = Enum("GameVersionEnum", {"latest": "latest"} | {str(v).replace(".", "_"): str(v) for v in GAME_VERSIONS})

@cache
def generate(api: ApiVersion, game_version: GameVersionEnum, table: Table) -> dict: # type: ignore
    if game_version == GameVersionEnum.latest:
        return generate(api, list(GameVersionEnum)[1], table)

    ver = GameVersion.from_string(game_version.value)
    return table.make_generator(ver).generate(api)

def precache_data():
    for game_version in GameVersionEnum:
        print(f">>> Precaching version {game_version.value}:", flush=True)

        for tb in Table.effective():
            for api in tb.spec.model.keys():
                print(f"> {tb.title} [v{api}]", flush=True)
                generate(api, game_version, tb)

        print(flush=True)


from pydantic import Field, PositiveFloat
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config, NonEmptyStr
from erdb.typing.models.item import Item
from erdb.typing.models.effect import Effect


@dataclass(config=dt_config())
class Talisman(Item):
    weight: PositiveFloat = Field(...,
        description="Weight of the Talisman",
        example=0.3
    )
    effects: list[Effect] = Field(...,
        description="Effects of the Talisman",
        # example provided by Effect model
    )
    conflicts: list[NonEmptyStr] = Field(...,
        description="Array of other talismans this one conflicts with.",
        example=["Crimson Amber Medallion", "Crimson Amber Medallion +1", "Crimson Amber Medallion +2"]
    )


import pytest

from erdb.typing.game_version import GameVersion, GameVersionRange


def _make(major: str, minor: str, patch: str) -> GameVersion:
    assert len(minor) >= 2, "Minor part must be at least 2 digits"
    return GameVersion(major, minor, patch, [int(major), int(minor), int(patch)])

@pytest.fixture(scope="module")
def base() -> GameVersion:
    return _make("1", "02", "3")

@pytest.mark.parametrize("left,right", [
    (["1", "02", "3"], ["1", "02", "3"]),
    (["1", "03", "3"], ["1", "03", "3"]),
    (["3", "03", "3"], ["3", "03", "3"]),
])
def test_eq(left: list[str], right: list[str]):
    assert _make(*left) == _make(*right)

@pytest.mark.parametrize("left,right", [
    (["1", "02", "4"], ["1", "02", "3"]),
    (["1", "04", "3"], ["1", "03", "3"]),
    (["4", "03", "3"], ["3", "03", "3"]),
])
def test_ne(left: list[str], right: list[str]):
    assert _make(*left) != _make(*right)

@pytest.mark.parametrize("left", [["1", "02", "2"], ["1", "01", "9"], ["0", "09", "9"]])
def test_lt(base, left):
    assert _make(*left) < base

@pytest.mark.parametrize("left", [["1", "02", "4"], ["1", "03", "0"], ["2", "00", "0"]])
def test_gt(base, left):
    assert _make(*left) > base

@pytest.mark.parametrize("left", [["1", "02", "2"], ["1", "02", "3"]])
def test_le(base, left):
    assert _make(*left) <= base

@pytest.mark.parametrize("left", [["1", "02", "3"], ["1", "02", "4"]])
def test_ge(base, left):
    assert _make(*left) >= base

@pytest.mark.parametrize("string,expected", [
    ("1.04.1", _make("1", "04", "1")),
    ("2.64.9", _make("2", "64", "9")),
    ("0.03.9",  _make("0", "03", "9")),
])
def test_from_string(string: str, expected: GameVersion):
    assert GameVersion.from_string(string) == expected

@pytest.mark.parametrize("string,begin,end", [
    ("from 1.04.1 until 1.05.0", _make("1", "04", "1"), _make("1", "05", "0")),
    ("from 1.04.1", _make("1", "04", "1"), GameVersion.max()),
    ("until 1.05.0", GameVersion.min(), _make("1", "05", "0")),
    ("only 1.04.1", _make("1", "04", "1"), GameVersion.max()),
    ("any version", GameVersion.min(), GameVersion.max()),
])
def test_range_from_string(string: str, begin: GameVersion, end: GameVersion):
    version_range = GameVersionRange.from_string(string)
    assert version_range.begin == begin
    assert version_range.end == end

@pytest.mark.parametrize("string,version", [
    ("from 1.04.1 until 1.05.0", _make("1", "04", "1")),
    ("from 1.04.1 until 1.05.0", _make("1", "04", "2")),
    ("from 1.04.1 until 1.05.0", _make("1", "04", "99")),
    ("from 1.04.1", _make("1", "04", "99")),
    ("from 1.04.1", _make("1", "05", "0")),
    ("until 1.05.0", _make("1", "04", "99")),
    ("until 1.05.0", _make("1", "03", "0")),
    ("only 1.05.0", _make("1", "05", "0")),
    ("any version", _make("1", "04", "99")),
    ("any version", _make("1", "05", "99")),
])
def test_range_contains(string: str, version: GameVersion):
    assert version in GameVersionRange.from_string(string)

@pytest.mark.parametrize("string,version", [
    ("from 1.04.1 until 1.05.0", _make("1", "04", "0")),
    ("from 1.04.1 until 1.05.0", _make("1", "05", "0")),
    ("from 1.04.1 until 1.05.0", _make("1", "05", "99")),
    ("from 1.04.1", _make("1", "04", "0")),
    ("from 1.04.1", _make("1", "03", "99")),
    ("until 1.05.0", _make("1", "05", "0")),
    ("only 1.05.0", _make("1", "04", "99")),
    ("only 1.05.0", _make("1", "05", "1")),
])
def test_range_not_contains(string: str, version: GameVersion):
    assert version not in GameVersionRange.from_string(string)


from pydantic import Field, NonNegativeInt, NonNegativeFloat, PositiveInt
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config
from erdb.typing.models.item import Item
from erdb.typing.models.common import Damage, StatRequirements
from erdb.typing.models.effect import Effect, StatusEffects
from erdb.typing.categories import ArmamentCategory
from erdb.typing.enums import Affinity, ArmamentUpgradeMaterial, AttackAttribute


@dataclass(config=dt_config())
class CorrectionCalcID:
    physical: NonNegativeInt
    magic: NonNegativeInt
    fire: NonNegativeInt
    lightning: NonNegativeInt
    holy: NonNegativeInt
    poison: NonNegativeInt
    bleed: NonNegativeInt
    sleep: NonNegativeInt
    madness: NonNegativeInt

@dataclass(config=dt_config())
class Scaling:
    strength: NonNegativeFloat | None = None
    dexterity: NonNegativeFloat | None = None
    intelligence: NonNegativeFloat | None = None
    faith: NonNegativeFloat | None = None
    arcane: NonNegativeFloat | None = None

@dataclass(config=dt_config())
class Guard:
    physical: NonNegativeInt | None = None
    magic: NonNegativeInt | None = None
    fire: NonNegativeInt | None = None
    lightning: NonNegativeInt | None = None
    holy: NonNegativeInt | None = None
    guard_boost: NonNegativeInt | None = None

@dataclass(config=dt_config())
class AffinityProperties:
    full_hex_id: str = Field(...,
        description="Full hex ID override for the Armament with Affinity applied.",
        regex=r"^[0-9A-F]+$",
        min_length=8, max_length=8,
        example="003085E0",
    )
    id: PositiveInt = Field(...,
        description="ID override for the Armament with Affinity applied.",
        exampe=3180000,
    )
    reinforcement_id: NonNegativeInt = Field(...,
        description="ID of reinforcement, refer to the `Reinforcements` table to look up value changes per level.",
        example=100,
    )
    correction_attack_id: NonNegativeInt = Field(...,
        description="ID of attack element correction, refer to the `Correction Attack` table to look up definitions.",
        example=10000,
    )
    correction_calc_id: CorrectionCalcID = Field(...,
        description="ID of calc correction for each damage type, refer to the `Correction Graph` table to look up value multipliers.",
        example=CorrectionCalcID(
            physical=0, magic=0, fire=0, lightning=0, holy=0,
            poison=6, bleed=6, sleep=6, madness=6,
        )
    )
    damage: Damage = Field(...,
        description="Base attack power values.",
        example=Damage(physical=138, stamina=63),
    )
    scaling: Scaling = Field(...,
        description="Base attribute scaling values.",
        example=Scaling(strength=0.49, dexterity=0.34)
    )
    guard: Guard = Field(...,
        description="Base guarded damage negation values.",
        example=Guard(physical=65, magic=35, fire=35, lightning=35, holy=35, guard_boost=42)
    )
    resistance: StatusEffects = Field(...,
        description="Base guarded resistances values.",
        example=StatusEffects(
            bleed=20, frostbite=20, poison=20, scarlet_rot=20,
            sleep=20, madness=20, death_blight=20
        ),
    )
    status_effects: StatusEffects = Field(...,
        description="Status effects of the Armament, dealt on hit.",
        example=StatusEffects(poison=80),
    )
    status_effect_overlay: list[StatusEffects] = Field(...,
        description="Array of status effects per Armament level which get upgraded alongside. Given the Armament level as the index to this array, the value should be overlaid on the `status_effects` property. In practice, only a single effect is set to be upgradable, but technically game params can do up to three.",
        max_items=26,
        example=[
            StatusEffects(poison=80), StatusEffects(poison=81), StatusEffects(poison=82),
            StatusEffects(poison=84), StatusEffects(poison=85), StatusEffects(poison=87),
            StatusEffects(poison=88), StatusEffects(poison=89), StatusEffects(poison=91),
            StatusEffects(poison=92), StatusEffects(poison=94), StatusEffects(poison=95),
            StatusEffects(poison=96), StatusEffects(poison=98), StatusEffects(poison=99),
            StatusEffects(poison=101), StatusEffects(poison=102), StatusEffects(poison=103),
            StatusEffects(poison=105), StatusEffects(poison=106), StatusEffects(poison=108),
            StatusEffects(poison=109), StatusEffects(poison=110), StatusEffects(poison=112),
            StatusEffects(poison=113), StatusEffects(poison=115),
        ]
    )

@dataclass(config=dt_config())
class Armament(Item):
    behavior_variation_id: NonNegativeInt = Field(...,
        description="Behavior variation ID used to identify attack params.",
        example=318
    )
    category: ArmamentCategory = Field(...,
        description="Class of the Armament.",
        example=ArmamentCategory.GREATSWORD,
    )
    weight: NonNegativeFloat = Field(...,
        description="Weight of the Armament.",
        example=9.,
    )
    default_skill_id: int = Field(...,
        description="Index of the default Skill the Armament comes with.",
        ge=10,
        example=100,
    )
    allow_ash_of_war: bool = Field(...,
        description="Specifies whether other Ashes of War can be put on the Armament and its affinity potentially changed.",
        example=True,
    )
    is_buffable: bool = Field(...,
        description="Specifies whether the Armament is buffable.",
        example=True,
    )
    is_l1_guard: bool = Field(...,
        description="Specifies whether the Armament is used for guarding when equipped in left hand.",
        example=True,
    )
    upgrade_material: ArmamentUpgradeMaterial = Field(...,
        description="Stones the Armament upgrades with, if upgradable.",
        example=ArmamentUpgradeMaterial.SMITHING_STONE,
    )
    upgrade_costs: list[NonNegativeInt] = Field(...,
        description="Array of Rune costs to upgrade to each level, +1 starting at position 0. Empty if the Armament is non-upgradable, otherwise it contains either 10 or 25 integers. `upgrade_material` can be used to tell the actual length.",
        min_items=0, max_items=25,
        example=[530, 636, 742, 848, 954, 1060, 1166, 1272, 1378, 1484],
    )
    attack_attributes: list[AttackAttribute] = Field(...,
        description="List of attack attributes the Armament can deal, usually 2.",
        min_items=1, max_items=2, unique_items=True,
        example=[AttackAttribute.PIERCE, AttackAttribute.STANDARD],
    )
    sp_consumption_rate: NonNegativeFloat = Field(...,
        description="Multiplier used for calculating the effective stamina consumption from the Skill's base stamina cost.",
        example=1.
    )
    requirements: StatRequirements = Field(...,
        description="Attribute requirements of the Armament.",
        example=StatRequirements(strength=16, dexterity=13),
    )
    effects: list[Effect] = Field(...,
        description="Effects of the Armament.",
        # example provided by Effect model
    )
    affinity: dict[Affinity, AffinityProperties] = Field(...,
        description="Mapping of possible affinities to their individual properties. `Standard` is always present.",
    )


import erdb.effect_parser.parsers as parse
from erdb.typing.effects import EffectModel, EffectType, AttributeName, AttributeField
from erdb.typing.enums import AttackCondition


_ATTRIBUTE_FIELDS = {
    "maxHpRate": AttributeField.create(
        attribute=AttributeName.MAXIMUM_HEALTH,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changeHpPoint": AttributeField.create(
        attribute=AttributeName.HEALTH_POINTS,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "changeHpRate": AttributeField.create(
        attribute=AttributeName.HEALTH_POINTS,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse_percentage,
        default_value=0
    ),
    "changeHpEstusFlaskCorrectRate": AttributeField.create(
        attribute=AttributeName.FLASK_HEALTH_RESTORATION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "maxMpRate": AttributeField.create(
        attribute=AttributeName.MAXIMUM_FOCUS,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changeMpPoint": AttributeField.create(
        attribute=AttributeName.FOCUS_POINTS,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "changeMpRate": AttributeField.create(
        attribute=AttributeName.FOCUS_POINTS,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse_percentage,
        default_value=0
    ),
    "changeMpEstusFlaskCorrectRate": AttributeField.create(
        attribute=AttributeName.FLASK_FOCUS_RESTORATION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "maxStaminaRate": AttributeField.create(
        attribute=AttributeName.MAXIMUM_STAMINA,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "staminaRecoverChangeSpeed": AttributeField.create(
        attribute=AttributeName.STAMINA_RECOVERY_SPEED,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "equipWeightChangeRate": AttributeField.create(
        attribute=AttributeName.MAXIMUM_EQUIP_LOAD,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "toughnessDamageCutRate": AttributeField.create(
        attribute=AttributeName.POISE,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.poise
    ),
    "addLifeForceStatus": AttributeField.create(
        attribute=AttributeName.VIGOR,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "addWillpowerStatus": AttributeField.create(
        attribute=AttributeName.MIND,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "addEndureStatus": AttributeField.create(
        attribute=AttributeName.ENDURANCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "addStrengthStatus": AttributeField.create(
        attribute=AttributeName.STRENGHT,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "addDexterityStatus": AttributeField.create(
        attribute=AttributeName.DEXTERITY,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "addMagicStatus": AttributeField.create(
        attribute=AttributeName.INTELLIGENCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "addFaithStatus": AttributeField.create(
        attribute=AttributeName.FAITH,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "addLuckStatus": AttributeField.create(
        attribute=AttributeName.ARCANE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "neutralDamageCutRate": AttributeField.create(
        attribute=AttributeName.STANDARD_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "blowDamageCutRate": AttributeField.create(
        attribute=AttributeName.STRIKE_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "slashDamageCutRate": AttributeField.create(
        attribute=AttributeName.SLASH_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "thrustDamageCutRate": AttributeField.create(
        attribute=AttributeName.PIERCE_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "defEnemyDmgCorrectRate_Physics": AttributeField.create(
        attribute=AttributeName.PHYSICAL_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "magicDamageCutRate": AttributeField.create(
        attribute=AttributeName.MAGIC_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "defEnemyDmgCorrectRate_Magic": AttributeField.create(
        attribute=AttributeName.MAGIC_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "fireDamageCutRate": AttributeField.create(
        attribute=AttributeName.FIRE_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "defEnemyDmgCorrectRate_Fire": AttributeField.create(
        attribute=AttributeName.FIRE_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "thunderDamageCutRate": AttributeField.create(
        attribute=AttributeName.LIGHTNING_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "defEnemyDmgCorrectRate_Thunder": AttributeField.create(
        attribute=AttributeName.LIGHTNING_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "darkDamageCutRate": AttributeField.create(
        attribute=AttributeName.HOLY_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "defEnemyDmgCorrectRate_Dark": AttributeField.create(
        attribute=AttributeName.HOLY_ABSORPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic_inverse
    ),
    "neutralAttackPowerRate": AttributeField.create(
        attribute=AttributeName.STANDARD_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "blowAttackPowerRate": AttributeField.create(
        attribute=AttributeName.STRIKE_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "slashAttackPowerRate": AttributeField.create(
        attribute=AttributeName.SLASH_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "thrustAttackPowerRate": AttributeField.create(
        attribute=AttributeName.PIERCE_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "physicsAttackPowerRate": AttributeField.create(
        attribute=AttributeName.PHYSICAL_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "physicsAttackRate": AttributeField.create(
        attribute=AttributeName.PHYSICAL_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "atkEnemyDmgCorrectRate_Physics": AttributeField.create(
        attribute=AttributeName.PHYSICAL_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "magicAttackRate": AttributeField.create(
        attribute=AttributeName.MAGIC_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "atkEnemyDmgCorrectRate_Magic": AttributeField.create(
        attribute=AttributeName.MAGIC_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "fireAttackRate": AttributeField.create(
        attribute=AttributeName.FIRE_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "atkEnemyDmgCorrectRate_Fire": AttributeField.create(
        attribute=AttributeName.FIRE_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "thunderAttackRate": AttributeField.create(
        attribute=AttributeName.LIGHTNING_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "atkEnemyDmgCorrectRate_Thunder": AttributeField.create(
        attribute=AttributeName.LIGHTNING_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "darkAttackRate": AttributeField.create(
        attribute=AttributeName.HOLY_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "atkEnemyDmgCorrectRate_Dark": AttributeField.create(
        attribute=AttributeName.HOLY_ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "staminaAttackRate": AttributeField.create(
        attribute=AttributeName.STAMINA_ATTACK_RATE,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "guardStaminaCutRate": AttributeField.create(
        attribute=AttributeName.STABILITY,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changePoisonResistPoint": AttributeField.create(
        attribute=AttributeName.POISON_RESISTANCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changeDiseaseResistPoint": AttributeField.create(
        attribute=AttributeName.SCARLET_ROT_RESISTANCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changeBloodResistPoint": AttributeField.create(
        attribute=AttributeName.BLEED_RESISTANCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changeFreezeResistPoint": AttributeField.create(
        attribute=AttributeName.FROSTBITE_RESISTANCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changeSleepResistPoint": AttributeField.create(
        attribute=AttributeName.SLEEP_RESISTANCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changeMadnessResistPoint": AttributeField.create(
        attribute=AttributeName.MADNESS_RESISTANCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changeCurseResistPoint": AttributeField.create(
        attribute=AttributeName.DEATH_BLIGHT_RESISTANCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "changeMagicSlot": AttributeField.create(
        attribute=AttributeName.MEMORY_SLOTS,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "dexterityCancelSystemOnlyAddDexterity": AttributeField.create(
        attribute=AttributeName.CASTING_SPEED,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "extendLifeRate": AttributeField.create(
        attribute=AttributeName.SPELL_DURATION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "magicConsumptionRate": AttributeField.create(
        attribute=AttributeName.SORCERY_FOCUS_CONSUMPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.NEGATIVE,
        parser=parse.generic
    ),
    "miracleConsumptionRate": AttributeField.create(
        attribute=AttributeName.INCANTATION_FOCUS_CONSUMPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.NEGATIVE,
        parser=parse.generic
    ),
    "shamanConsumptionRate": AttributeField.create( # likely unused
        attribute=AttributeName.PYROMANCY_FOCUS_CONSUMPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.NEGATIVE,
        parser=parse.generic
    ),
    "artsConsumptionRate": AttributeField.create(
        attribute=AttributeName.SKILL_FOCUS_CONSUMPTION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.NEGATIVE,
        parser=parse.generic
    ),
    "bowDistRate": AttributeField.create(
        attribute=AttributeName.BOW_DISTANCE,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "hearingSearchEnemyRate": AttributeField.create(
        attribute=AttributeName.ENEMY_HEARING,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.NEGATIVE,
        parser=parse.generic
    ),
    "fallDamageRate": AttributeField.create(
        attribute=AttributeName.FALL_DAMAGE,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.NEGATIVE,
        parser=parse.generic
    ),
    "itemDropRate": AttributeField.create(
        attribute=AttributeName.ITEM_DISCOVERY,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.item_discovery
    ),
    "soulRate": AttributeField.create(
        attribute=AttributeName.RUNE_ACQUISITION,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
    "soul": AttributeField.create(
        attribute=AttributeName.RUNE_ACQUISITION,
        effect_model=EffectModel.ADDITIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic
    ),
}

_WEAPON_ATTRIBUTE_FIELDS = {
    "weakA_DamageRate": AttributeField.create(
        attribute=AttributeName.ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic,
        conditions=[str(AttackCondition.VS_GRAVITY_ENEMIES)]
    ),
    "weakB_DamageRate": AttributeField.create(
        attribute=AttributeName.ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic,
        conditions=[str(AttackCondition.VS_UNDEAD_ENEMIES)]
    ),
    "weakC_DamageRate": AttributeField.create(
        attribute=AttributeName.ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic,
        conditions=[str(AttackCondition.VS_DRAGON_ENEMIES)]
    ),
    "weakD_DamageRate": AttributeField.create(
        attribute=AttributeName.ATTACK_POWER,
        effect_model=EffectModel.MULTIPLICATIVE,
        effect_type=EffectType.POSITIVE,
        parser=parse.generic,
        conditions=[str(AttackCondition.VS_ANCIENT_DRAGON_ENEMIES)]
    ),
}

def get(weapon: bool = False) -> dict[str, AttributeField]:
    return _WEAPON_ATTRIBUTE_FIELDS if weapon else _ATTRIBUTE_FIELDS


from erdb.typing.params import ParamDict
from erdb.shop.shop_typing import Material, Lineup, Product


"""
Helper class for looking up any sort of item exchanges (purchases, alterations, crafting...)
"""
class Lookup(object):
    _shop_lineup: ParamDict
    _material_sets: ParamDict

    def __init__(self, shop_lineup: ParamDict, material_sets: ParamDict) -> None:
        self._shop_lineup = shop_lineup
        self._material_sets = material_sets

    def get_lineups_from_material(self, material: Material) -> list[Lineup]:
        lineups: list[Lineup] = []

        for lineup_param in self._shop_lineup.values():
            if mat_id := lineup_param["mtrlId"].get_int():
                lineup = Lineup.from_params(lineup_param, self._material_sets[mat_id])
                if material in lineup.materials.keys():
                    lineups.append(lineup)

        return lineups


from collections import defaultdict

from erdb.loaders.params import load as load_params
from erdb.typing.params import ParamDict
from erdb.typing.enums import ItemIDFlag
from erdb.typing.game_version import GameVersion


def _get_values(effects: ParamDict, field: str, limit: int = 10) -> dict[str, list[str]]:
    values = defaultdict(list)

    for effect in effects.values():
        val = effect[field]
        case = str(effect.index) if len(effect.name) == 0 else effect.name

        if len(values[val]) < limit:
            values[val].append(case)

        elif len(values[val]) == limit:
            values[val].append("(...)")

    return values

def find_valid_values(param_name: str, version: str, field: str, limit: int = 8):
    """
    Interesting param.fields overview:
    * SpEffectParam:conditionHp                      -- trigger when HP below %
    * SpEffectParam:conditionHpRate                  -- trigger when HP above %
    * SpEffectParam:invocationConditionsStateChange1 -- seemingly the only differentiating field for Concealing Veil
    * SpEffectParam:invocationConditionsStateChange1/2/3
    * SpEffectParam:toughnessDamageCutRate           -- kinda like inverted poise
    * SpEffectParam:miracleConsumptionRate           -- FP consumption rate for incantations
    * SpEffectParam:shamanConsumptionRate            -- FP consumption rate for pyormancies (?)
    * SpEffectParam:magicConsumptionRate             -- FP consumption rate for sorceries
    * SpEffectParam:stateInfo                        -- a lot of unique effects seem to use this and this only
    * SpEffectParam:guardStaminaCutRate              -- Greatshield Talisman doesn't seem to use this
    * SpEffectParam:magicSubCategoryChange1/2/3      -- seem to specify conditions exclusively
    """

    params = load_params(param_name, GameVersion.from_string(version), ItemIDFlag.NON_EQUIPABBLE)
    values = _get_values(params, field, limit=8 if limit < 0 else limit)

    for value in sorted(values.keys(), key=float):
        cases = values[value]
        value = f"{value}".ljust(8)
        print(f">> {value}:", ', '.join(cases))


from typing import Iterator
import xml.etree.ElementTree as xmltree
from csv import DictReader
from zipfile import Path as ZipPath

from erdb.loaders import PKG_DATA_PATH
from erdb.typing.game_version import GameVersion
from erdb.typing.params import ParamRow, ParamDict
from erdb.typing.enums import ItemIDFlag


def _load(param: str, version: GameVersion) -> Iterator[dict[str, str]]:
    archive = PKG_DATA_PATH / "gamedata" / f"{version}.zip"
    with ZipPath(archive, at=f"{param}.csv").open(mode="r") as f:
        yield from DictReader(f, delimiter=";") # type: ignore

def load(param: str, version: GameVersion, item_id_flag: ItemIDFlag) -> ParamDict:
    return {int(row["Row ID"]): ParamRow.make(row, item_id_flag) for row in _load(param, version)}

# optimal variant for params with a lot of IDs like spEffects
def load_ids(param: str, version: GameVersion, item_id_flag: ItemIDFlag, id_min: int, id_max: int = 999999999) -> ParamDict:
    return {int(row["Row ID"]): ParamRow.make(row, item_id_flag) for row in _load(param, version) if id_min <= int(row["Row ID"]) <= id_max}

def load_msg(filename: str, version: GameVersion) -> dict[int, str]:
    archive = PKG_DATA_PATH / "gamedata" / f"{version}.zip"
    with ZipPath(archive, at=f"{filename}.fmg.xml").open(mode="r", encoding="utf-8") as f:
        tree = xmltree.fromstring(f.read())
        entries = tree.findall(".//text")

    return {int(str(e.get("id"))): str(e.text) for e in entries if e.text != "%null%"}


from typing import Any, NamedTuple, Self, overload
from erdb.typing.enums import ItemIDFlag


class ParamField(str):
    @property
    def as_str(self) -> str:
        return self

    @property
    def as_int(self) -> int:
        return int(self)

    @property
    def as_bool(self) -> bool:
        return self != "0"

    @property
    def as_float(self) -> float:
        return float(self)

    @overload
    def get_int(self, default: int, null_value: Any = "-1", formatter = lambda x: x) -> int: ...

    @overload
    def get_int(self, default: int | None = None, null_value: Any = "-1", formatter = lambda x: x) -> int | None: ...

    def get_int(self, default: int | None = None, null_value: Any = "-1", formatter = lambda x: x) -> int | None:
        return default if self == str(null_value) else formatter(int(self))

    @overload
    def get_float(self, default: float, null_value: Any = "-1", formatter = lambda x: x) -> float: ...

    @overload
    def get_float(self, default: float | None = None, null_value: Any = "-1", formatter = lambda x: x) -> float | None: ...

    def get_float(self, default: float | None = None, null_value: Any = "-1", formatter = lambda x: x) -> float | None:
        return default if self == str(null_value) else formatter(float(self))

class ParamRow(NamedTuple):
    index: int
    item_id_flag: ItemIDFlag
    name: str
    field_dict: dict[str, str]

    @property
    def index_hex(self) -> str:
        assert self.item_id_flag != ItemIDFlag.DISABLE_CHECK
        return f"{self.index + self.item_id_flag:08X}"

    @property
    def is_base_item(self) -> bool:
        """
        Retrieves whether the item is a non-upgraded version of itself. These usually differ by last two digits of the index.
        """
        return self.index % 100 == 0

    def __getitem__(self, key: str) -> ParamField:
        assert key in self.field_dict, f"\"{key}\" not found"
        return ParamField(self.field_dict[key])

    def __contains__(self, __x: object) -> bool:
        return __x in self.field_dict

    @classmethod
    def make(cls, field_dict: dict[str, str], item_id_flag: ItemIDFlag) -> Self:
        index = int(field_dict["Row ID"])
        return cls(
            index=index,
            item_id_flag=item_id_flag,
            name=field_dict["Row Name"],
            field_dict=field_dict,
        )

ParamDict = dict[int, ParamRow]


import json
from enum import Enum
from pathlib import Path
from time import sleep


try:
    import gspread
except ModuleNotFoundError:
    assert False, "Module \"gspread\" is not installed, please install it manually"

class Field(Enum):
    CLASS = "G10"
    NAME = "G11"
    AFFINITY = "G12"
    LEVEL = "G13"

    STRENGTH = "G15"
    DEXTERITY = "G16"
    INTELLIGENCE = "G17"
    FAITH = "G18"
    ARCANE = "G19"

    AP_PHYSICAL_BASE = "G30"
    AP_PHYSICAL_SCALING = "J30"
    AP_PHYSICAL_ROUNDED = "L30" # rounded

    AP_MAGIC_BASE = "G31"
    AP_MAGIC_SCALING = "J31"
    AP_MAGIC_ROUNDED = "L31" # rounded

    AP_FIRE_BASE = "G32"
    AP_FIRE_SCALING = "J32"
    AP_FIRE_ROUNDED = "L32" # rounded

    AP_LIGHTNING_BASE = "G33"
    AP_LIGHTNING_SCALING = "J33"
    AP_LIGHTNING_ROUNDED = "L33" # rounded

    AP_HOLY_BASE = "G34"
    AP_HOLY_SCALING = "J34"
    AP_HOLY_ROUNDED = "L34" # rounded

    SB_PHYSICAL_BASE = "G35"
    SB_PHYSICAL_SCALING = "J35"
    SB_PHYSICAL_ROUNDED = "L35" # rounded

    SB_MAGIC_BASE = "G36"
    SB_MAGIC_SCALING = "J36"
    SB_MAGIC_ROUNDED = "L36" # rounded

    SB_FIRE_BASE = "G37"
    SB_FIRE_SCALING = "J37"
    SB_FIRE_ROUNDED = "L37" # rounded

    SB_LIGHTNING_BASE = "G38"
    SB_LIGHTNING_SCALING = "J38"
    SB_LIGHTNING_ROUNDED = "L38" # rounded

    SB_HOLY_BASE = "G39"
    SB_HOLY_SCALING = "J39"
    SB_HOLY_ROUNDED = "L39" # rounded

    POISON_BASE = "G40"
    POISON_SCALING = "J40"
    POISON_ROUNDED = "L40" # rounded

    SCARLET_ROT_BASE = "G41"
    SCARLET_ROT_SCALING = "J41"
    SCARLET_ROT_ROUNDED = "L41" # rounded

    BLEED_BASE = "G42"
    BLEED_SCALING = "J42"
    BLEED_ROUNDED = "L42" # rounded

    FROSTBITE_BASE = "G43"
    FROSTBITE_SCALING = "J43"
    FROSTBITE_ROUNDED = "L43" # rounded

    SLEEP_BASE = "G44"
    SLEEP_SCALING = "J44"
    SLEEP_ROUNDED = "L44" # rounded

    MADNESS_BASE = "G45"
    MADNESS_SCALING = "J45"
    MADNESS_ROUNDED = "L45" # rounded

    @staticmethod
    def attribs() -> list["Field"]:
        return [Field.STRENGTH, Field.DEXTERITY, Field.INTELLIGENCE, Field.FAITH, Field.ARCANE]

_ATTRIB_SET: list[str] = [
    "20,18,30,30,7",  # holy
    "9,7,30,30,20",   # holy unoptimal
    "20,80,11,15,7",  # dexterity
    "8,80,9,9,20",    # dexterity unoptimal
    "30,18,11,70,7",  # faith
    "9,7,8,70,20",    # faith unoptimal
    "20,18,60,15,7",  # intelligence
    "9,7,60,9,20",    # intelligence unoptimal
    "1,1,1,1,1",      # lowest
    "20,18,11,15,80", # arcane
    "40,40,11,15,7",  # quality
    "40,40,6,8,20",   # quality unoptimal
    "10,10,10,10,10", # starting wretch
    "14,13,9,9,7",    # starting vagabond
    "60,18,11,15,7",  # strength
    "60,12,9,9,20",   # strength unoptimal
    "99,99,99,99,99", # unscaled shitter
]

def retry_on_quota_limit(times: int, time: int):
    def _decorator(func):
        def _newfn(*args, **kwargs):
            attempt = 0

            while attempt < times:
                try:
                    return func(*args, **kwargs)

                except gspread.exceptions.APIError as e:
                    print(f"Quota limit likely exceeded, trying again in {time} second(s)...", flush=True)
                    sleep(time)
                    attempt += 1

            return func(*args, **kwargs)

        return _newfn
    return _decorator

def _get_ingame_name(name: str) -> str:
    return {
        "Misericorde": "Miséricorde",
        "Great Epee": "Great Épée",
        "Varre's Bouquet": "Varré's Bouquet",
    }.get(name, name)

class OnlineCalc:
    _sheet: gspread.Worksheet
    _writes: int
    _reads: int

    def __init__(self, version: str, service_key: Path) -> None:
        gc = gspread.service_account(filename=service_key)
        self._sheet = gc.open(f"{version} Weapon AP Calculator").worksheet("Sheet1")
        self._writes = 0
        self._reads = 0

    @property
    def writes(self) -> int:
        return self._writes

    @property
    def reads(self) -> int:
        return self._reads

    def _inc_write(self):
        self._writes += 1

    def _inc_read(self):
        self._reads += 1

    @retry_on_quota_limit(times=60, time=5)
    def set_field(self, field: Field, value):
        self._sheet.update(field.value, value)
        self._inc_write()

    @retry_on_quota_limit(times=60, time=5)
    def set_fields(self, fields: list[Field], values: list):
        assert len(fields) == len(values)
        batch = [{"range": f.value, "values": [[v]]} for f, v in zip(fields, values)]
        self._sheet.batch_update(batch)
        self._inc_write()

    @retry_on_quota_limit(times=60, time=5)
    def get_field(self, field: Field):
        value = self._sheet.acell(field.value).value
        self._inc_read()
        return value

    @retry_on_quota_limit(times=60, time=5)
    def get_fields(self, start: Field, end: Field):
        values = self._sheet.get(f"{start.value}:{end.value}")
        self._inc_read()
        return values

    def set_armament(self, class_: str, name: str):
        # cannot self.set_fields because every field is checked individually
        self.set_field(Field.CLASS, class_)
        self.set_field(Field.NAME, name)

    def set_attribs(self, attribs: str):
        parts = attribs.split(",")
        assert len(parts) == 5

        self.set_fields(Field.attribs(), [*map(int, parts)])

def _get_attack_values(ap: list[str]) -> list[float]:
    return [float(v) for v in ap if v not in ["", "-   "]]

"""
gspread requires a path to the service key, downloaded from Google API when
setting up a serivce account login. Detailed instructions here:
https://docs.gspread.org/en/latest/oauth2.html#for-bots-using-service-account
"""
def fetch(version: str, service_key: str):
    calc = OnlineCalc(version, Path(service_key))
    calc.set_field(Field.LEVEL, 10) # always keep 10, valid for unique and standard armaments

    with open(Path.cwd() / version / "armaments.json", encoding="utf-8") as f:
        armaments = json.load(f)["Armaments"]

    # take sample for testing
    #armaments = {k: armaments[k] for k in list(armaments)[:30]}

    for attribs in _ATTRIB_SET:
        print(f"Attributes \"{attribs}\"", flush=True)

        calc.set_attribs(attribs)
        results = dict()

        count = len(armaments)

        for armament, data in armaments.items():
            print(f"Armament \"{armament}\" ({count} remaining)", flush=True)
            count -= 1

            print(f"Reads: {calc.reads}, writes: {calc.writes}", flush=True)

            calc.set_armament(data["class"], _get_ingame_name(armament))
            results[armament] = dict()

            for affinity in data["affinity"].keys():
                print(f"Affinity \"{affinity}\"", flush=True)

                calc.set_field(Field.AFFINITY, affinity)
                values = calc.get_fields(Field.AP_PHYSICAL_BASE, Field.MADNESS_ROUNDED)

                results[armament][affinity] = {
                    "attack_power": {
                        "physical": _get_attack_values(values[0]),
                        "magic": _get_attack_values(values[1]),
                        "fire": _get_attack_values(values[2]),
                        "lightning": _get_attack_values(values[3]),
                        "holy": _get_attack_values(values[4]),
                    },
                    "spell_buff": {
                        "physical": _get_attack_values(values[5]),
                        "magic": _get_attack_values(values[6]),
                        "fire": _get_attack_values(values[7]),
                        "lightning": _get_attack_values(values[8]),
                        "holy": _get_attack_values(values[9]),
                    },
                    "status_effects": {
                        "poison": _get_attack_values(values[10]),
                        "scarlet_rot": _get_attack_values(values[11]),
                        "bleed": _get_attack_values(values[12]),
                        "frostbite": _get_attack_values(values[13]),
                        "sleep": _get_attack_values(values[14]),
                        "madness": _get_attack_values(values[15]),
                    }
                }

        with open(Path.cwd() / f"{version}-attacks-{attribs}.json", mode="w") as f:
            json.dump(results, f, indent=4)

    print(f"Finished in {calc.reads} reads and {calc.writes} writes.", flush=True)



import erdb.effect_parser.attribute_fields as attrib_fields
import erdb.effect_parser.parsers as parse
import erdb.effect_parser.hardcoded as hardcoded_effects
from erdb.effect_parser.aggregator import aggregate_effects
from erdb.typing.models.effect import StatusEffects
from erdb.typing.params import ParamRow, ParamDict
from erdb.typing.enums import SpEffectType, AttackCondition
from erdb.typing.effects import SchemaEffect


_REFERENCE_EFFECT_PARAMS: list[str] = ["cycleOccurrenceSpEffectId", "applyIdOnGetSoul"]

_SP_EFFECT_TO_FIELD: dict[SpEffectType, str] = {
    SpEffectType.HEMORRHAGE: "bloodAttackPower",
    SpEffectType.FROSTBITE: "freezeAttackPower",
    SpEffectType.POISON: "poizonAttackPower",
    SpEffectType.SCARLET_ROT: "diseaseAttackPower",
    SpEffectType.SLEEP: "sleepAttackPower",
    SpEffectType.MADNESS: "madnessAttackPower",
    SpEffectType.BLIGHT: "curseAttackPower",
}

_SP_EFFECT_TO_STR: dict[SpEffectType, str] = {
    SpEffectType.HEMORRHAGE: "bleed",
    SpEffectType.FROSTBITE: "frostbite",
    SpEffectType.POISON: "poison",
    SpEffectType.SCARLET_ROT: "scarlet_rot",
    SpEffectType.SLEEP: "sleep",
    SpEffectType.MADNESS: "madness",
    SpEffectType.BLIGHT: "death_blight",
}

def get_effects(sp_effect: ParamRow, sp_effect_type: SpEffectType, triggeree: ParamRow | None = None, init_conditions: list[str] | None = None) -> list[SchemaEffect]:
    effects = hardcoded_effects.get(sp_effect.index, sp_effect_type)

    for field, attrib_field in attrib_fields.get().items():
        if sp_effect[field] == str(attrib_field.default_value):
            continue

        effect = SchemaEffect.from_attribute_field(sp_effect[field].as_float, attrib_field)

        effect.conditions = init_conditions
        if conds := parse.conditions(sp_effect, triggeree):
            effect.conditions = conds if effect.conditions is None else effect.conditions + conds

        effect.tick_interval = parse.interval(sp_effect)
        effect.value_pvp = parse.value_pvp(sp_effect, field, attrib_fields.get())

        effects.append(effect)

    return effects

def get_effects_nested(sp_effect: ParamRow, sp_effects: ParamDict, add_condition: AttackCondition | None) -> list[SchemaEffect]:
    sp_effect_type = SpEffectType(sp_effect["stateInfo"])
    effects = get_effects(sp_effect, sp_effect_type, init_conditions=[str(add_condition)] if add_condition else None)

    for ref_id in (sp_effect[ref_field].as_int for ref_field in _REFERENCE_EFFECT_PARAMS):
        if ref_sp_effect := sp_effects.get(ref_id):
            if ref_sp_effect.index > 0:
                effects += get_effects(ref_sp_effect, sp_effect_type, sp_effect)

    for condition_offset in hardcoded_effects.get_conditions(sp_effect.index):
        ref_sp_effect = sp_effects[sp_effect.index + condition_offset.offset]
        init_conditions = None if condition_offset.condition is None else [str(condition_offset.condition)]
        effects += get_effects(ref_sp_effect, sp_effect_type, sp_effect, init_conditions)

    return effects

def get_status_effect(sp_effect: ParamRow) -> tuple[str, int]:
    # NOTE: not identifying effects by values, relying on `stateInfo` to be correct at all times
    etype = SpEffectType(sp_effect["stateInfo"])
    return _SP_EFFECT_TO_STR[etype], sp_effect[_SP_EFFECT_TO_FIELD[etype]].as_int

def parse_effects(row: ParamRow, sp_effects: ParamDict, *effect_referencing_fields: str, add_condition: AttackCondition | None = None) -> list[dict]:
    effects: list[SchemaEffect] = []

    for effect_id in (row[ref_field].as_int for ref_field in effect_referencing_fields):
        if effect_id in hardcoded_effects.get_status_effect_ranges():
            continue

        if effect_id in sp_effects:
            effects += get_effects_nested(sp_effects[effect_id], sp_effects, add_condition)

    return [e.to_dict() for e in aggregate_effects(effects)]

def parse_status_effects(effect_ids: list[int], sp_effects: ParamDict) -> StatusEffects:
    # Getting 0th effect if value no found, bug with Antspur Rapier -- get anything to return a 0 status effect
    effects = [sp_effects.get(i, sp_effects[0]) for i in effect_ids if i != -1]
    status_effects = hardcoded_effects.get_status_effect_ranges()
    return StatusEffects(**dict([get_status_effect(e) for e in effects if e.index in status_effects]))

def parse_weapon_effects(weapon: ParamRow) -> list[dict]:
    effects: list[SchemaEffect] = []

    for field, attrib_field in attrib_fields.get(weapon=True).items():
        if weapon[field] != str(attrib_field.default_value):
            effects.append(SchemaEffect.from_attribute_field(weapon[field].as_float, attrib_field))

    return [e.to_dict() for e in effects]